library CH341A_8255;

// fr ProfiLab Expert Software ab V4.0 neuestes Update
// (www.abacom-online.de)

uses
  SysUtils,
  Windows,
  Classes,
  Controls,
  Dlg_Config in 'Dlg_Config.pas' {ConfigDlg},
  CH341DLL in 'CH341DLL.PAS'; {API DLL von abacom}

{$R *.RES}

Const
       // USER VARIABLES STORED IN PUSER^
       DevNo = 0; // ch341 device number
       PortStatusOld = 1; // latest output status
       PortSel = 2; // 8255 port selection A,B,CL,CH
       PortDir = 3; // 8255 port direction
       ChipSel = 4; // Chip Select
       TimeLastCall = 5; // Latest execution time of Calculate EX
       ClockMode = 6; // Clock AUTO or from USER (from CLK input)
       ClockOldStatus = 7; // Old clock status

       // ClockModes
       ClockModeAuto = 0;
       ClockModeUser = 1;

       // Port selection constants
       PortASel  = 0;
       PortBSel  = 1;
       PortCLSel = 2;
       PortCHSel = 3;

       // 8255 port direction constants
       PortDirInput = 0;
       PortDirOutput = 1;
       PortDirOnlyRead = 2; //only read port; no direction setting

       // 8255 Register addresses
       PortAReg = 0;
       PortBReg = 1;
       PortCReg = 2;
       ControlReg = 3;

       // Maximum constants
       Max8255 = 63;  // 64x 8255 addressable?
       MaxCH341A = 100; // 100x ch341A on USB?


Type TDLLParams = array[0..100] of extended;   //Type of ProfiLab DLL parameters
     PDLLParams = ^TDLLParams;                 // Pointer to ProfiLab DLL parameters


// Global Variables
var ControlRegs: array[0..Max8255, 0..MaxCh341A] of Byte; //max. 64pcs. 8255 chips on 100 pcs. CH341A addressable - Oh well!
    PortChar: String[2]; // makes nice pin names
    ClockChannel: Integer; // Remeber ClockChannel;

// ******** 8255 control funtions **********

procedure WriteToAddressLatch(DevNo, aByte: Byte);
var WriteLen: Cardinal;
begin
   WriteLen:=1;
   ch341MemWriteAddr0(DevNo, @aByte, @WriteLen);
end;

procedure WriteToDataPort(DevNo, aByte: Byte);
var WriteLen: Cardinal;
begin
   WriteLen:=1;
   ch341MemWriteAddr1(DevNo, @aByte, @WriteLen);
end;

function ReadDataPort(DevNo: Byte): Byte;
var ReadLen: Cardinal;
    aByte: Byte;
begin
   ReadLen:=1;
   ch341MemReadAddr1(DevNo, @aByte, @ReadLen);
   result:=aByte;
end;

Procedure SetControlReg(PUser: PDLLParams);
var Device: Byte;
    Adr: Byte;
    ControlByte: Byte;
begin

   // ControlRegs = Global array of control regs for all possible 8255!
   ControlByte:=ControlRegs[round(PUser^[ChipSel]), Device];

   //Init ControlReg: all ports as input, select Mode0, if not done so before
   if ControlByte and $80=0 then ControlByte:=$9B;

// 8255 Control Register

// Port direction A, B, CH, CL (in Bit 4,3,1,0)
// Bit clear = Output port
// Bit set = Input port

// Mode 0 is fixed (in Bit 7,6,5,2)

// Bit 7  6  5  4  3  2  1  0
//     1  0  0  A  CH 0  B CL

   case round(PUser^[PortDir]) of
   PortDirOutput: //Clear bits -> Dir = Output
   begin
      case round(PUser^[PortSel]) of
      PortASel: ControlByte:=ControlByte and (not (1 shl 4));
      PortBSel: ControlByte:=ControlByte and (not (1 shl 1));
      PortCLSel: ControlByte:=ControlByte and (not (1 shl 0));
      PortCHSel: ControlByte:=ControlByte and (not (1 shl 3));
      end;
   end;
   PortDirInput: //Set bits -> Dir = Input
   begin
      case round(PUser^[PortSel]) of
      PortASel: ControlByte:=ControlByte or (1 shl 4);
      PortBSel: ControlByte:=ControlByte or (1 shl 1);
      PortCLSel: ControlByte:=ControlByte or (1 shl 0);
      PortCHSel: ControlByte:=ControlByte or (1 shl 3);
      end;
   end;
   end;
   // Store back to global
   ControlRegs[round(PUser^[ChipSel]), Device]:=ControlByte;

   // and finaly write to device
   case round(PUser^[PortDir]) of
   PortDirOnlyRead: ; // nichts; nur lesen
   PortDirOutput, PortDirInput:
   begin
      //CH341A device
      Device:=round(PUser^[DevNo]);

      // Chip Select + Register Select
      Adr:=round(PUser^[ChipSel]) shl 2;
      Adr:=Adr+ControlReg;
      WriteToAddressLatch(Device, Adr);

      WriteToDataPort(Device, ControlByte);
   end;
   end;
end;

//**************** ProfiLab DLL functions *************************

function NumInputsEx(PUser: PDLLParams): Byte;
begin
  case round(PUser^[PortDir]) of
  PortDirOutput:
     case round(PUser^[PortSel]) of
     PortASel:
     begin
        result:= 8;
        PortChar:='A';
     end;
     PortBSel:
     begin
        result:= 8;
        PortChar:='B';
     end;
     PortCLSel:
     begin
        result:= 4;
        PortChar:='CL';
     end;
     PortCHSel:
     begin
        result:= 4;
        PortChar:='CH';
     end;
     end;
  else
     result:=0;
  end;
  if round(PUser^[ClockMode])=ClockModeUser then
  begin
     // USER CLK -> add CLK pin
     ClockChannel:=result;
     result:=result+1;
  end
  else
  begin
     // AUTO CLK -> no clock pin
     ClockChannel:=-1;
  end;
end;

function NumOutputsEx(PUser: PDLLParams): Byte;
begin
  case round(PUser^[PortDir]) of
  PortDirInput, PortDirOnlyRead:
     case round(PUser^[PortSel]) of
     PortASel:
     begin
        result:= 8;
        PortChar:='A';
     end;
     PortBSel:
     begin
        result:= 8;
        PortChar:='B';
     end;
     PortCLSel:
     begin
        result:= 4;
        PortChar:='CL';
     end;
     PortCHSel:
     begin
        result:= 4;
        PortChar:='CH';
     end;
     end;
  else
     result:=0;
  end;
end;

Function InputName(Channel: Byte): ShortString; // Return name for each component input pin
begin
   if Channel=ClockChannel then
   begin
      result:='|CLK';
   end
   else // normal channels...
   begin
      if PortChar='CH' then
      begin
         result:='P'+PortChar[1]+ intToStr(Channel+4);
      end
      else
      begin
         result:='P'+PortChar[1]+ intToStr(Channel);
      end;
   end;
end;

Function OutputName(Channel: Byte): ShortString; // Return name for each component output pin
begin
   if PortChar='CH' then
   begin
      result:='P'+PortChar[1]+ intToStr(Channel+4);
   end
   else
   begin
      result:='P'+PortChar[1]+ intToStr(Channel);
   end;
end;

Procedure SimStart(PInput,POutput,PUser: PDLLParams); //called when ProfiLab enters RUN mode
var Device: Integer;
begin
  Device:=round(PUser^[DevNo]);  // CH341A device no.
  CH341OpenDevice(Device);       // Verbinden mit CH341A
  SetControlReg(PUser);          // Set port direction
  PUser^[PortStatusOld]:=-1;     // Make sure output state is set in CalculateEx first time called
end;

Procedure SimStop(PInput,POutput,PUser: PDLLParams); //called when ProfiLab RUN mode is terminated
var Device: Integer;
begin
  //CH341A device no.
  Device:=round(PUser^[DevNo]);
  CH341CloseDevice(round(PUser^[DevNo])); //Verbinung schliessen
end;

//called regularly from ProfiLab
Procedure Calculate(PInput,POutput,PUser: PDLLParams);
var i: Integer;
    l: Cardinal;
    Buffer: Byte;
    Device: Byte;
    Adr: Byte;
    PortStatus: Byte;
    NumBits: Byte;
    TimeNow: DWord;
    CLK: Boolean;
    CLK_old: Boolean;
    Triggered: Boolean;
    ClkMode: Integer;
begin

   ClkMode:=round(PUser^[ClockMode]); // Get Clock Mode
   Triggered:=false;

   if ClkMode=ClockModeUser then
   begin
      // Clock mode: USER
      ClockChannel:=NumInputsEx(PUser)-1; //Get Clock Channel
      CLK:=(PInput^[ClockChannel] > 2.5);  //Get CLK input
      CLK_old:=(PUser^[ClockOldStatus] > 2.5);  //Get old clock Status
      if CLK<>CLK_old then // Clock input changed
      begin
         if CLK then PUser^[ClockOldStatus]:=5 else PUser^[ClockOldStatus]:=0; //remember for next time
         if not CLK then Triggered:= true; //trigger on falling edge
      end;
      if not Triggered then exit; // Not triggered -> nothing to do; return to PL immediately
   end
   else
   begin
      // Clock mode: AUTO
      // limit sample rate to 20ms; keeps Profilab alive
      TimeNow:=GetTickCount;
      if TimeNow<round(PUser^[TimeLastCall]) then PUser^[TimeLastCall]:=TimeNow;
      if TimeNow-round(PUser^[TimeLastCall])<20 then exit;
      PUser^[TimeLastCall]:=TimeNow;
   end;


   case round(PUser^[PortDir]) of // 8255 port direction

   PortDirInput, PortDirOnlyRead:
   begin
      //CH341A device no.
      Device:=round(PUser^[DevNo]);

      // Chip Select + Register Select
      Adr:=round(PUser^[ChipSel]) shl 2; // Chip Select with upper adress bits (A2..A7)
      case round(PUser^[PortSel]) of     // Register Select with lower address bits (A0, A1)
      PortASel: Adr:=Adr+PortAReg;
      PortBSel: Adr:=Adr+PortBReg;
      PortCLSel: Adr:=Adr+PortCReg;
      PortCHSel: Adr:=Adr+PortCReg;
      end;
      WriteToAddressLatch(Device, Adr);

      //read port
      PortStatus:=ReadDataPort(Device);
      if round(PUser^[PortSel])=PortCHSel then PortStatus:=PortStatus shr 4;
      NumBits:=NumOutputsEx(PUser);
      For i:=0 to Numbits-1 do
      begin
         if PortStatus and (1 shl i)>0 then POutput^[i]:=5 else POutput^[i]:=0;
      end;
   end;

   PortDirOutput:
   begin
      PortStatus:=0;
      NumBits:=NumInputsEx(PUser);
      if ClkMode=ClockModeUser then Numbits:=NumBits-1; // One less; Last is CLK input!

      //Get port status...
      For i:=0 to Numbits-1 do
      begin
         if PInput^[i]>2.5 then PortStatus:=PortStatus or (1 shl i);
      end;
      if round(PUser^[PortSel])=PortCHSel then PortStatus:=PortStatus shl 4;

      // Communication triggered by port status change in clock mode 'AUTO'
      if ClkMode=ClockModeAuto then Triggered:=(PortStatus <> round(PUser^[PortStatusOld]));

      if Triggered then //Do communication....
      begin
         //CH341A device no.
         Device:=round(PUser^[DevNo]);

         // Chip Select + Register Select
         Adr:=round(PUser^[ChipSel]) shl 2; // Chip Select with upper adress bits (A2..A7)
         case round(PUser^[PortSel]) of     // Register Select with lower address bits (A0, A1)
         PortASel: Adr:=Adr+PortAReg;
         PortBSel: Adr:=Adr+PortBReg;
         PortCLSel: Adr:=Adr+PortCReg;
         PortCHSel: Adr:=Adr+PortCReg;
         end;
         WriteToAddressLatch(Device, Adr);

         // Write port data
         WriteToDataport(Device,PortStatus);
         PUser^[PortStatusOld]:=PortStatus;
      end;
   end;

   end;
end;


Procedure Configure(PUser: PDLLParams);
var DLG: TConfigDlg;
begin
   //Create user dialogue
   DLG:=TConfigDlg.Create(nil);
   With DLG do
   begin
      // Dialogfelder vorbelegen
      se_DevNo.Value:=round(PUser^[DevNo]); //CH341A device no.
      se_ChipSel.Value:=round(PUser^[ChipSel]); //8255 Chip Select
      cb_PortSel.itemIndex:=round(PUser^[PortSel]); //8255 Port Select
      cb_PortDir.itemIndex:=round(PUser^[PortDir]); //8255 Port Direction
      cb_ClockMode.itemIndex:=round(PUser^[ClockMode]); //Clock Mode
      //Dialog ausfhren und bei OK..
      if ShowModal=mrOK then
      begin
         // Dialogfelder auslesen
         PUser^[DevNo]:=se_DevNo.Value;
         PUser^[ChipSel]:=se_ChipSel.Value;
         PUser^[PortSel]:=cb_PortSel.ItemIndex;
         PUser^[PortDir]:=cb_PortDir.ItemIndex;
         PUser^[ClockMode]:=cb_ClockMode.ItemIndex;
      end;
      // Free dialogue
      free;
   end;
end;

//export methods for ProfiLab
exports SimStart,
        SimStop,
        NumInputsEx,
        NumOutputsEx,
        Calculate,
        InputName,
        OutputName,
        Configure;
begin
end.

